//! Acme client that supports multiple CAs and configs for sites that can be seperate from the mainconfig
#![allow(clippy::clone_on_copy)]
#![allow(clippy::identity_op)]
#![allow(refining_impl_trait)]
#![allow(clippy::collapsible_if)]
#![allow(dead_code)]

pub(crate) mod consts;
pub(crate) mod macros;
pub(crate) mod prelude;
pub(crate) mod process;
pub(crate) mod types;
pub(crate) mod utils;

#[cfg(feature = "capabilities")]
use crate::utils::check_permissions;
use crate::{
    consts::*,
    prelude::*,
    types::{
        config::{
            GeneralConfig,
            SiteConfig,
        },
        dns::Manager,
        structs::{
            Arguments,
            Error,
            ProcessorArgs,
            SubCommand,
        },
    },
};
use acme2_eab::Directory;
use async_scoped::TokioScope;
use clap::Parser;
use env_logger::init as log_init;
use libsystemd::daemon;
use log::*;
use openssl::{
    self,
    pkey::{
        PKey,
        Private,
    },
};
use reqwest::tls::Version;
use schemars::{
    SchemaGenerator,
    consts::meta_schemas::DRAFT07,
    generate::SchemaSettings,
};
use serde::Serialize;
use serde_json::ser::{
    Formatter,
    PrettyFormatter,
};
use std::{
    collections::{
        HashMap,
        HashSet,
    },
    path::{
        Path,
        PathBuf,
    },
    process::exit,
    str::FromStr,
    sync::Arc,
    time::Duration,
};
use tokio::{
    fs::{
        create_dir_all,
        read_dir,
    },
    io::{
        AsyncReadExt,
        AsyncWriteExt,
    },
    sync::Mutex,
};
use tokio_stream::{
    StreamExt,
    wrappers::ReadDirStream,
};


fn default_client() -> Result<reqwest::Client, Error> {
    reqwest::Client::builder()
        .min_tls_version(Version::TLS_1_2)
        .https_only(true)
        .pool_max_idle_per_host(POOL_SIZE)
        .build()
        .map_err(Error::from_display)
}

async fn load_privkey(path: PathBuf) -> Result<PKey<Private>, Error> {
    let mut file = match FILE_MODE.open(path).await {
        Ok(file) => file,
        Err(error) => return Error::err(format!("Failed to open Private Key: {error}")),
    };
    let mut data = String::new();
    if let Err(error) = file.read_to_string(&mut data).await {
        return Error::err(format!("Failed to read data for the key: {error}"));
    }
    match PKey::private_key_from_pem(data.as_bytes()) {
        Ok(key) => Ok(key),
        Err(error) => Error::err(format!("Failed to parse pem data: {error}")),
    }
}

async fn racme(flags: Arguments) -> Result<(), Error> {
    let client = default_client()?;
    let dns_manager = Manager::new();
    let systemd_access = daemon::booted();
    let mainconfig = {
        let file = match FILE_MODE.open(flags.config).await {
            Ok(file) => file,
            Err(error) => {
                return Error::err(format!("error reading the config: {error}"));
            },
        };
        GeneralConfig::from_file(file).await
    };
    trace!("Parsed Config: {mainconfig:?}");
    let files = {
        let rd = match read_dir(mainconfig.sites_path.clone()).await {
            Ok(rd) => rd,
            Err(error) => {
                return Error::err(format!("could not read files from sites dir: {error}"));
            },
        };
        ReadDirStream::new(rd)
    };
    let mut siteconfigs = Vec::new();
    for filename in files.filter(Result::is_ok).map(|file| file.unwrap().path()).collect::<Vec<PathBuf>>().await {
        let file = match FILE_MODE.open(filename.clone()).await {
            Ok(file) => file,
            Err(error) => {
                warn!("Failed to read the configfile {}: {}", filename.display(), error);
                continue;
            },
        };
        let mut site = SiteConfig::from_file(file).await;
        site.name = filename.file_stem().unwrap().to_string_lossy().to_string();
        siteconfigs.push(site);
    }
    let used = siteconfigs.iter().map(|s| s.ca.clone()).collect::<HashSet<_>>();
    debug!("Used CAs: {used:?}");

    let mut directories = HashMap::<String, Arc<Directory>>::new();

    let mut accounts = HashMap::new();
    let accountpath = Path::new(&mainconfig.accounts_path).to_path_buf();
    if let Err(error) = create_dir_all(accountpath.clone()).await {
        return Error::err(format!("Failed to create the directory for the accounts: {error}"));
    }
    for (name, ca) in mainconfig.ca.iter().filter(|(name, _)| used.contains(name.to_owned())) {
        process::accounts(name, ca, &mut directories, &mut accounts, &client, accountpath.clone()).await;
    }

    let restart_services = Mutex::new(HashSet::<String>::new());
    let reload_services = Mutex::new(HashSet::<String>::new());
    let certs = Path::new(&mainconfig.certificates_path).to_path_buf();

    if !certs.exists() {
        if let Err(error) = create_dir_all(certs.clone()).await {
            return Error::err(format!("Failed to create directory for all the certificates: {error}"));
        }
    }
    let challengepath = mainconfig.http_challenge_path.and_then(|path| PathBuf::from_str(path.as_str()).ok());

    unsafe {
        TokioScope::scope_and_collect(|scope| {
            for site in siteconfigs {
                if let Some(account) = accounts.get(&site.ca) {
                    scope.spawn(process::site(ProcessorArgs::new(
                        site,
                        Arc::clone(account),
                        &reload_services,
                        &restart_services,
                        certs.clone(),
                        challengepath.clone(),
                        dns_manager.clone(),
                        client.clone(),
                    )));
                } else {
                    error!("Could not process site {} because of previous errors", site.name)
                }
            }
        })
    }
    .await;

    if systemd_access {
        process::services(restart_services.into_inner(), reload_services.into_inner()).await;
    }
    Ok(())
}

fn serialize_with_formatter<T: Serialize, F: Formatter>(value: &T, formatter: F) -> Result<String, Error> {
    let mut store = Vec::with_capacity(2 ^ 10);
    let mut serializer = serde_json::ser::Serializer::with_formatter(&mut store, formatter);
    match value.serialize(&mut serializer) {
        Ok(_) => {},
        Err(error) => return Error::err(format!("Failed to Serialize the schema: {error}")),
    }
    store.push(b'\n');
    Ok(unsafe { String::from_utf8_unchecked(store) })
}

async fn schema_generator() -> Result<(), Error> {
    let formatter = PrettyFormatter::with_indent(&[b' '; 4]);
    let mut schema_settings = SchemaSettings::default();
    schema_settings.meta_schema = Some(DRAFT07.into());
    let mut generator = SchemaGenerator::new(schema_settings);
    let general_schema = serialize_with_formatter(&generator.root_schema_for::<GeneralConfig>(), formatter.clone())?;
    match FILE_MODE_WRITE.clone().create_new(false).open("schema-general.json").await {
        Ok(mut file) => {
            match file.write(general_schema.as_bytes()).await {
                Ok(_) => {},
                Err(error) => return Error::err(format!("{error}")),
            }
        },
        Err(error) => return Err(Error::from_display(error)),
    };

    let site_schema = serialize_with_formatter(&generator.root_schema_for::<SiteConfig>(), formatter.clone())?;
    match FILE_MODE_WRITE.clone().create_new(false).open("schema-site.json").await {
        Ok(mut file) => {
            match file.write(site_schema.as_bytes()).await {
                Ok(_) => {},
                Err(error) => return Error::err(format!("{error}")),
            }
        },
        Err(error) => return Err(Error::from_display(error)),
    };
    Ok(())
}

fn main() {
    log_init();
    let args = Arguments::parse();
    #[cfg(feature = "capabilities")]
    if args.subcommands.is_none() && !check_permissions() {
        error!(
            "This program needs the capability to change the ownership and the permissions of files. this can be done via adding the capability via `capsh --caps=\"cap_chown+ep cap_fowner+ep\" --shell=racme -- racme.toml`, systemd service setting AmbientCapabilities or running as root(not recommended)"
        );
        exit(4)
    }
    let runtime = match tokio::runtime::Builder::new_current_thread().enable_all().build() {
        Ok(runtime) => runtime,
        Err(error) => {
            error!("Could not initialize Tokio runtime: {error}");
            exit(2)
        },
    };
    let result = match args.subcommands {
        None => runtime.block_on(racme(args)),
        Some(SubCommand::Schema) => runtime.block_on(schema_generator()),
    };
    runtime.shutdown_timeout(Duration::from_secs(1));
    if let Err(error) = result {
        error!("{error}");
        exit(6)
    }
}
